"""Entrypoints to the nwp-consumer service."""

import logging
import os
import sys
from typing import NamedTuple

from nwp_consumer.internal import handlers, ports, repositories

log = logging.getLogger("nwp-consumer")

class Adaptors(NamedTuple):
    """Adaptors for the CLI."""
    model_repository: type[ports.RawRepository]
    notification_repository: type[ports.NotificationRepository]

def parse_env() -> Adaptors:
    """Parse from the environment."""
    model_repository_adaptor: type[ports.RawRepository]
    match os.getenv("MODEL_REPOSITORY"):
        # Default to NOAA S3 as it is freely accessible
        case None | "gfs":
            model_repository_adaptor = \
                repositories.raw_repositories.NOAAS3RawRepository
        case "ceda":
            model_repository_adaptor = \
                repositories.raw_repositories.CEDARawRepository
        case "ecmwf-realtime":
            model_repository_adaptor = \
                repositories.raw_repositories.ECMWFRealTimeS3RawRepository
        case "metoffice-datahub":
            model_repository_adaptor = \
                repositories.raw_repositories.MetOfficeDatahubRawRepository
        case "ecmwf-mars":
            model_repository_adaptor = \
                    repositories.raw_repositories.ECMWFMARSRawRepository
        case _ as mr:
            log.error(
                f"Unknown model repository '{mr}'. Expected one of "
                f"['gfs', 'ceda', 'ecmwf-realtime', 'metoffice-datahub', 'ecmwf-mars']",
            )
            sys.exit(1)

    notification_repository_adaptor: type[ports.NotificationRepository]
    match os.getenv("NOTIFICATION_REPOSITORY", "stdout"):
        case "stdout":
            notification_repository_adaptor = \
                repositories.notification_repositories.StdoutNotificationRepository
        case "dagster-pipes":
            notification_repository_adaptor = \
                repositories.notification_repositories.DagsterPipesNotificationRepository
        case _ as notification:
            log.error(f"Unknown notification repository: {notification}")
            sys.exit(1)

    return Adaptors(
        model_repository=model_repository_adaptor,
        notification_repository=notification_repository_adaptor,
    )


def run_cli() -> None:
    """Entrypoint for the CLI handler."""
    adaptors = parse_env()
    c = handlers.CLIHandler(
        model_adaptor=adaptors.model_repository,
        notification_adaptor=adaptors.notification_repository,
    )
    returncode: int = c.run()
    sys.exit(returncode)




"""Model repository implementation for data from ECMWF's MARS catalogue.

Repository Information
======================

Todo:
Documented structure
--------------------
"""

import dataclasses
import datetime as dt
import inspect
import logging
import os
import pathlib
from collections.abc import Callable, Iterator
from typing import override

import cfgrib
import xarray as xr
from ecmwfapi import ECMWFService
from joblib import delayed
from returns.result import Failure, ResultE, Success

from nwp_consumer.internal import entities, ports

log = logging.getLogger("nwp-consumer")


def _mars_logger(msg: str) -> None:
    """Redirect log from ECMWF API to structlog.

    Keyword Arguments:
    ------------------
    msg: The message to redirect.
    """
    debugSubstrings: list[str] = ["Requesting", "Transfering", "efficiency", "Done"]
    errorSubstrings: list[str] = ["ERROR", "FATAL"]
    if any(map(msg.__contains__, debugSubstrings)):
        log.debug("[MARS] %s", msg)
    if any(map(msg.__contains__, errorSubstrings)):
        log.warning("[MARS] %s", msg)

@dataclasses.dataclass
class _MARSRequest:
    """A request for data in the MARS format.

    See Also:
        - https://confluence.ecmwf.int/display/UDOC/MARS+request+syntax
    """

    params: list[entities.Parameter]
    """The parameters to request."""
    init_time: dt.datetime
    """The init time of the data."""
    steps: list[int]
    """The forecast steps."""
    nwse: str
    """The area of the grid to request.

    The format is 'N/W/S/E'.
    """

    classfication: str = "od"
    """The ECMWF classification given to the data."""
    expver: int = 1
    """The version of the data. 1 is 'Operational'."""
    levtype: str = "sfc"
    """The type of level."""
    stream: str = "oper"
    """The forecasting system used to generate the data.

    Options include:
        - "oper" for operational
        - "enfo" for ensemble forecast
        - "enda" for ensemble data assimilation
    """
    field_type: str = "fc"
    """Determines the type of the field to be retrieved.

    Options include:
        - "fc" for forecast
        - "em" for ensemble mean
        - "es" for ensemble standard deviation
        - "pf" for perturbed forecast (full ensemble)
    """
    grid: str = "0.1/0.1"
    """The grid resolution."""
    number: list[int] | None = None
    """The ensemble member numbers to request.

    Only relevant in full ensemble data requests.
    """

    def as_ensemble_mean_request(self) -> "_MARSRequest":
        """Create a new request for the ensemble mean."""
        return dataclasses.replace(
            self,
            field_type="em",
            stream="enfo",
        )

    def as_ensemble_std_request(self) -> "_MARSRequest":
        """Create a new request for the ensemble standard deviation."""
        return dataclasses.replace(
            self,
            field_type="es",
            stream="enfo",
        )

    def as_full_ensemble_request(self) -> "_MARSRequest":
        """Create a new request for the full ensemble."""
        return dataclasses.replace(
            self,
            field_type="pf",
            stream="enfo",
        )

    def as_operational_request(self) -> "_MARSRequest":
        """Create a new request for the operational data."""
        return dataclasses.replace(
            self,
            field_type="fc",
            stream="oper",
        )

    def gen_filename(self) -> str:
        """Generate a filename for the request.

        The stream and type are encoded into the filename, as it is required information
        at conversion time. This is because the data does not have the ensemble_stat
        dimension, so the type must be encoded in the filename to be added in.
        """
        return f"ecmwf_{self.stream}-{self.field_type}_{self.init_time:%Y%m%dT%H}.grib"

    def _to_string(self, method: str, target: str) -> str:
        """Make a MARS-formatted request string.

        Args:
            method: The method to use. Either 'list' or 'retrieve'.
            target: Path to the file where the data will be stored.
        """
        param: str = "/".join([p.metadata().grib2_code for p in self.params])
        step: str = "/".join(map(str, self.steps))

        marsReq: str = f"""
            {"list" if method == "list" else "retrieve"},
                class = {self.classfication},
                date = {self.init_time:%Y%m%d},
                expver = {self.expver},
                levtype = {self.levtype},
                stream = {self.stream},
                param = {param},
                step = {step},
                time = {self.init_time:%H},
        """
        marsReq += f"number = {'/'.join(map(str, self.number))}," if self.number is not None else ""
        marsReq += f"""
                type = {self.field_type},
                area = {self.nwse},
                grid = {self.grid},
                target = "{target}"
        """
        return inspect.cleandoc(marsReq)

    def execute(self, server: ECMWFService, folder: pathlib.Path) -> ResultE[pathlib.Path]:
        """Execute the request on the server.

        Args:
            server: The server to execute the request on.
            folder: The folder to store the data in.

        Returns:
            ResultE with path to the downloaded file.
        """
        target: pathlib.Path = folder / self.gen_filename()
        # TODO: Check against a MARS LIST call first?
        try:
            log.debug("Downloading to '%s'", target)
            server.execute(
                self._to_string(method="retrieve", target=target.as_posix()),
                target=target.as_posix(),
            )
            log.debug("Downloaded to '%s'", target)
        except Exception as e:
            return Failure(OSError(
                "Failed to download data from ECMWF MARS. "
                "Ensure request targets available parameters and steps. "
                f"Error context: {e}",
            ))
        return Success(target)


class ECMWFMARSRawRepository(ports.RawRepository):
    """Model repository implementation for archive data from ECMWF's MARS."""

    server: ECMWFService

    def __init__(self, server: ECMWFService) -> None:
        """Create a new instance of the class."""
        self.server = server

    @staticmethod
    @override
    def repository() -> entities.RawRepositoryMetadata:
        return entities.RawRepositoryMetadata(
            name="ECMWF-MARS",
            is_archive=True,
            is_order_based=False,
            delay_minutes=(60 * 26),  # 1 day, plus leeway
            max_connections=20,
            required_env=[
                "ECMWF_API_KEY",
                "ECMWF_API_EMAIL",
                "ECMWF_API_URL",
            ],
            optional_env={},
            postprocess_options=entities.PostProcessOptions(),
            available_models={
                "default": entities.Models.ECMWF_HRES_IFS_0P1DEGREE.with_region("uk"),
                "hres-ifs-uk": entities.Models.ECMWF_HRES_IFS_0P1DEGREE.with_region("uk"),
                "hres-ifs-india": entities.Models.ECMWF_HRES_IFS_0P1DEGREE.with_region("india"),
                "hres-ifs-west-europe": entities.Models.ECMWF_HRES_IFS_0P1DEGREE\
                    .with_region("west-europe"),
                "ens-stat-india": entities.Models.ECMWF_ENS_STAT_0P1DEGREE.with_region("india"),
                "ens-stat-uk": entities.Models.ECMWF_ENS_STAT_0P1DEGREE.with_region("uk"),
                "ens-uk": entities.Models.ECMWF_ENS_0P1DEGREE\
                    .with_region("uk")\
                    .with_chunk_count_overrides({"latitude": 1, "longitude": 1}),
            },
        )

    @staticmethod
    @override
    def model() -> entities.ModelMetadata:
        requested_model: str = os.getenv("MODEL", default="default")
        if requested_model not in ECMWFMARSRawRepository.repository().available_models:
            log.warn(
                f"Unknown model '{requested_model}' requested, falling back to default ",
                "ECMWF-MARS repository only supports "
                f"'{list(ECMWFMARSRawRepository.repository().available_models.keys())}'. "
                "Ensure MODEL environment variable is set to a valid model name.",
            )
            requested_model = "default"
        return ECMWFMARSRawRepository.repository().available_models[requested_model]

    @classmethod
    @override
    def authenticate(cls) -> ResultE["ECMWFMARSRawRepository"]:
        missing_envs = cls.repository().missing_required_envs()
        if len(missing_envs) > 0:
            return Failure(OSError(
                "Cannot authenticate with ECMWF's MARS service due to "
                f"missing required environment variables: {', '.join(missing_envs)}",
            ))
        # Auth is picked up from required environment variables
        server = ECMWFService(
            service="mars",
        )
        return Success(cls(server=server))

    @override
    def fetch_init_data(
        self, it: dt.datetime,
    ) -> Iterator[Callable[..., ResultE[list[xr.DataArray]]]]:
        req: _MARSRequest = _MARSRequest(
            params=self.model().expected_coordinates.variable,
            init_time=it,
            steps=self.model().expected_coordinates.step,
            nwse="/".join([str(ord) for ord in self.model().expected_coordinates.nwse()]),
            number=self.model().expected_coordinates.ensemble_member,
        )

        # Yield the download and convert function with the appropriate request type
        if self.model().expected_coordinates.ensemble_stat is not None:
            for stat_req in [req.as_ensemble_mean_request(), req.as_ensemble_std_request()]:
                yield delayed(self._download_and_convert)(stat_req)
        elif self.model().expected_coordinates.ensemble_member is not None:
            yield delayed(self._download_and_convert)(req.as_full_ensemble_request())
        else:
            yield delayed(self._download_and_convert)(req.as_operational_request())

    def _download_and_convert(self, mr: _MARSRequest) -> ResultE[list[xr.DataArray]]:
        """Download and convert data from the ECMWF MARS server.

        Args:
            mr: The request to download data from.
        """
        return self._download(mr).bind(self._convert)

    def _download(self, mr: _MARSRequest) -> ResultE[pathlib.Path]:
        """Download data from the ECMWF MARS server.

        Args:
            mr: The request to download data from.
        """
        local_folder: pathlib.Path = (
            pathlib.Path(
                os.getenv(
                    "RAWDIR",
                    f"~/.local/cache/nwp/{self.repository().name}/{self.model().name}/raw",
                ),
            )
        ).expanduser()
        local_folder.mkdir(parents=True, exist_ok=True)

        local_path: pathlib.Path = local_folder / mr.gen_filename()
        if local_path.exists():
            return Success(local_path)

        return mr.execute(server=self.server, folder=local_folder)

    @staticmethod
    def _convert(path: pathlib.Path) -> ResultE[list[xr.DataArray]]:
        """Convert a local grib file to xarray DataArrays.

        Args:
            path: The path to the file to convert.
        """
        try:
            dss: list[xr.Dataset] = cfgrib.open_datasets(
                path=path.as_posix(),
                chunks={"time": 1, "step": -1, "longitude": -1, "latitude": -1},
                backend_kwargs={"indexpath": ""},
            )
        except Exception as e:
            return Failure(OSError(
                f"Failed to convert raw MARS data at '{path!s}' to xarray. "
                "Ensure file is in GRIB format and contains expected data. "
                f"Error context: {e}",
            ))

        processed_das: list[xr.DataArray] = []
        try:
            # Merge the datasets back into one
            ds: xr.Dataset = xr.merge(
                objects=dss,
                compat="override",
                combine_attrs="drop_conflicts",
            )
            del dss

            # Add in missing coordinates for mean/std data
            # * I don't really like basing this off the file name
            # * TODO: Find a better way
            if "enfo-es" in path.name:
                ds = ds.expand_dims(dim={"ensemble_stat": ["std"]})
            elif "enfo-em" in path.name:
                ds = ds.expand_dims(dim={"ensemble_stat": ["mean"]})
            da: xr.DataArray = (
                ds
                .pipe(
                    entities.Parameter.rename_else_drop_ds_vars,
                    allowed_parameters=ECMWFMARSRawRepository.model().expected_coordinates.variable,
                )
                .rename({"time": "init_time"})
                .expand_dims("init_time")
                .to_dataarray(name=ECMWFMARSRawRepository.model().name)
            )
            if "enfo-pf" in path.as_posix():
                da = da.rename({"number": "ensemble_member"})
            da = (
                da.drop_vars(
                    names=[
                        c for c in ds.coords
                        if c not in ECMWFMARSRawRepository.model().expected_coordinates.dims
                    ],
                    errors="ignore",
                )
                .transpose(*ECMWFMARSRawRepository.model().expected_coordinates.dims)
                .sortby(variables=["step", "variable", "longitude"])
                .sortby(variables="latitude", ascending=False)
            )
            # Put each variable into its own DataArray:
            # * Each raw file does not contain a full set of parameters
            # * and so may not produce a contiguous subset of the expected coordinates.
            processed_das.extend(
                [
                    da.where(cond=da.coords["variable"] == v, drop=True)
                    for v in da.coords["variable"].values
                ],
            )


        except Exception as e:
            return Failure(ValueError(
                f"Error processing DataArray for path '{path!s}'. Error context: {e}",
            ))

        return Success(processed_das)




import dataclasses
import datetime as dt
import unittest

from .postprocess import PostProcessOptions
from .repometadata import RawRepositoryMetadata


class TestRawRepositoryMetadata(unittest.TestCase):
    """Test the business methods of the RawRepositoryMetadata class."""

    metadata: RawRepositoryMetadata = RawRepositoryMetadata(
        name="test",
        is_archive=False,
        is_order_based=False,
        delay_minutes=55,
        required_env=["TEST"],
        optional_env={"TEST": "test"},
        max_connections=1,
        postprocess_options=PostProcessOptions(),
        available_models={},
    )

    def test_determine_latest_it_from(self) -> None:
        """Test the determine_latest_it_from method."""

        @dataclasses.dataclass
        class TestCase:
            name: str
            t: dt.datetime
            expected: dt.datetime

        tests = [
            TestCase(
                name="rolls_back_inter-day",
                t=dt.datetime(2021, 1, 2, 0, tzinfo=dt.UTC),
                expected=dt.datetime(2021, 1, 1, 18, tzinfo=dt.UTC),
            ),
            TestCase(
                name="rolls_back_intra-day",
                t=dt.datetime(2021, 1, 1, 5, tzinfo=dt.UTC),
                expected=dt.datetime(2021, 1, 1, 0, tzinfo=dt.UTC),
            ),
            TestCase(
                name="returns_to_hours",
                t=dt.datetime(2021, 1, 1, 6, tzinfo=dt.UTC),
                expected=dt.datetime(2021, 1, 1, 0, tzinfo=dt.UTC),
            ),
        ]

        for test in tests:
            with self.subTest(name=test.name):
                result = self.metadata.determine_latest_it_from(test.t, [0, 6, 12, 18])
                self.assertEqual(result, test.expected)


if __name__ == "__main__":
    unittest.main()



"""Struct definitions for domain entities.

These define data objects and behaviours that are used in the services core.

Domain Entities
---------------

Entities are the core building blocks of the domain layer. They are the
representations of the business objects that are manipulated by the application.

By using domain entities in the core, it is ensured that the business logic is
separated from the technical details of the application.

A domain entity may have associated methods that define its behaviour, but it
should not contain any logic that is specific to a particular implementation.
"""

from .repometadata import RawRepositoryMetadata
from .modelmetadata import ModelMetadata, Models
from .tensorstore import ParameterScanResult, TensorStore
from .postprocess import PostProcessOptions, CodecOptions
from .notification import PerformanceMetadata, StoreCreatedNotification, StoreAppendedNotification
from .parameters import Parameter
from .coordinates import NWPDimensionCoordinateMap
from .performance import PerformanceMonitor

__all__ = [
    "RawRepositoryMetadata",
    "ModelMetadata",
    "Models",
    "ParameterScanResult",
    "TensorStore",
    "PostProcessOptions",
    "CodecOptions",
    "PerformanceMetadata",
    "StoreCreatedNotification",
    "StoreAppendedNotification",
    "Parameter",
    "NWPDimensionCoordinateMap",
    "PerformanceMonitor",
]



"""Domain entities for NWP parameters.

NWP forecasts have to forecast something, and that something is the value
of one or more parameters. Often referred to as variables (or
channels once stored in a tensor), these parameters are physical, measurable
quantities that are forecasted by the model.

Variables are forecasted at different levels in the atmosphere. A common
level of interest is called "screen level (sl)". This corresponds to 1.5-2m
above the Earth's surface, which is the height of many measuring stations.
Other levels include units of distance in metres from the ground,
including a 0m level which is the ground itself (gl). Also, the mean sea
level (msl) is used as a reference point for pressure.

Variables also have units, which are the physical quantities that the
variable is measured in. For example, temperature is measured in degrees
Celsius (C), and wind speed is measured in meters per second (m/s).
Some variables just occupy the range [0, 1], such as cloud cover.
This unit is referred to as the Unit Interval (UI).

See Also:
    - https://datahub.metoffice.gov.uk/docs/glossary
    - https://codes.ecmwf.int/grib/param-db
"""

import dataclasses
import logging
from enum import StrEnum, auto

import xarray as xr
from returns.result import Failure, ResultE, Success

log = logging.getLogger("nwp-consumer")


@dataclasses.dataclass(slots=True)
class ParameterLimits:
    """Class containing information about the limits of a parameter."""

    upper: float
    """The upper limit on the parameter value.

    Not an absolute maximum, but rather the maximum value that
    the parameter can resonably be expected to take.

    As an example, the maximum distance that can be seen horizontally
    is 4.5km at sea level, so the upper limit for a visibility
    parameter should be ~4500m.
    """

    lower: float
    """The lower limit on the parameter value.

    Not an absolute minimum, but rather the minimum value that
    the parameter can reasonably be expected to take.

    As an example, the minimum temperature on Earth is -89C,
    so the lower limit for a temperature parameter should be ~ -90C.
    """

    threshold: float = 0.05
    """How carefully to enforce the limits.

    The threshold defines the number of values that can be outside
    the limits in a given dataset before the data is considered invalid.
    Denoted as a ratio (#outside/#total).
    """


@dataclasses.dataclass(slots=True, frozen=True)
class ParameterData:
    """Class containing information about a parameter."""

    name: str
    """The name of the parameter as appears in produced datasets."""

    description: str
    """A brief description of the parameter."""

    units: str
    """The units of the parameter."""

    limits: ParameterLimits
    """Reasonable physical limits for the parameter.

    Used in sanity and validity checking the database values.
    """

    grib2_code: str
    """The GRIB2 code for the parameter.

    See Also:
        - https://codes.ecmwf.int/grib/param-db/?filter=All
    """

    alternative_shortnames: list[str] = dataclasses.field(default_factory=list)
    """Alternate names for the parameter found in the wild."""


    def __str__(self) -> str:
        """String representation of the parameter."""
        return self.name


class Parameter(StrEnum):
    """Parameters of interest to OCF.

    Inheriting from StrEnum and using ``auto()`` makes the values
    of the enums equal to the lowercased enum name.

    See Also:
        - https://docs.python.org/3/library/enum.html#enum.StrEnum
    """

    TEMPERATURE_SL = auto()
    DOWNWARD_SHORTWAVE_RADIATION_FLUX_GL = auto()
    DOWNWARD_LONGWAVE_RADIATION_FLUX_GL = auto()
    RELATIVE_HUMIDITY_SL = auto()
    VISIBILITY_SL = auto()
    WIND_U_COMPONENT_10m = auto()
    WIND_V_COMPONENT_10m = auto()
    WIND_U_COMPONENT_100m = auto()
    WIND_V_COMPONENT_100m = auto()
    WIND_U_COMPONENT_200m = auto()
    WIND_V_COMPONENT_200m = auto()
    SNOW_DEPTH_GL = auto()
    CLOUD_COVER_HIGH = auto()
    CLOUD_COVER_MEDIUM = auto()
    CLOUD_COVER_LOW = auto()
    CLOUD_COVER_TOTAL = auto()
    TOTAL_PRECIPITATION_RATE_GL = auto()
    DOWNWARD_ULTRAVIOLET_RADIATION_FLUX_GL = auto()
    DIRECT_SHORTWAVE_RADIATION_FLUX_GL = auto()
    WIND_SPEED_10m = auto()
    WIND_SPEED_100m = auto()
    WIND_DIRECTION_10m = auto()
    PRESSURE_MSL = auto()

    def metadata(self) -> ParameterData:
        """Get the metadata for the parameter."""
        match self.name:

            case self.TEMPERATURE_SL.name:
                return ParameterData(
                    name=str(self),
                    description="Temperature at screen level",
                    units="C",
                    limits=ParameterLimits(upper=60, lower=-90),
                    alternative_shortnames=["t", "t2m", "tas"],
                    grib2_code="167",
                )

            case self.DOWNWARD_SHORTWAVE_RADIATION_FLUX_GL.name:
                return ParameterData(
                    name=str(self),
                    description="Downward shortwave radiation flux at ground level. "
                                "Defined as the mean amount of solar radiation "
                                "incident on the surface expected over the next hour."
                                "This is made up of both direct and diffuse radiation.",
                    units="W/m^2",
                    limits=ParameterLimits(upper=1500, lower=0),
                    alternative_shortnames=["swavr", "ssrd", "dswrf", "sdswrf"],
                    grib2_code="169",
                )

            case self.DOWNWARD_LONGWAVE_RADIATION_FLUX_GL.name:
                return ParameterData(
                    name=str(self),
                    description="Downward longwave radiation flux at ground level. "
                                "Defined as the mean amount of thermal radiation "
                                "incident on the surface expected over the next hour.",
                    units="W/m^2",
                    limits=ParameterLimits(upper=500, lower=0),
                    alternative_shortnames=["strd", "dlwrf", "sdlwrf"],
                    grib2_code="175",
                )

            case self.RELATIVE_HUMIDITY_SL.name:
                return ParameterData(
                    name=str(self),
                    description="Relative humidity at screen level. "
                                "Defined as the ratio of partial pressure of water vapour "
                                "to the equilibrium vapour pressure of water",
                    units="%",
                    limits=ParameterLimits(upper=100, lower=0),
                    alternative_shortnames=["r", "r2"],
                    grib2_code="157",
                )

            case self.VISIBILITY_SL.name:
                return ParameterData(
                    name=str(self),
                    description="Visibility at screen level. "
                                "Defined as the distance at which an object can be seen "
                                "horizontally in daylight conditions.",
                    units="m",
                    limits=ParameterLimits(upper=4500, lower=0),
                    alternative_shortnames=["vis"],
                    grib2_code="20",
                )

            case self.WIND_U_COMPONENT_10m.name:
                return ParameterData(
                    name=str(self),
                    description="U component of wind at 10m above ground level. "
                                "Defined as the horizontal speed of "
                                "the wind in the eastward direction.",
                    units="m/s",
                    limits=ParameterLimits(upper=100, lower=-100),
                    alternative_shortnames=["u10", "u", "uas"],
                    grib2_code="165",
                )

            case self.WIND_V_COMPONENT_10m.name:
                return ParameterData(
                    name=str(self),
                    description="V component of wind at 10m above ground level. "
                                "Defined as the horizontal speed of "
                                "the wind in the northward direction.",
                    units="m/s",
                    # Non-tornadic winds are usually < 100m/s
                    limits=ParameterLimits(upper=100, lower=-100),
                    alternative_shortnames=["v10", "v", "vas"],
                    grib2_code="166",
                )

            case self.WIND_U_COMPONENT_100m.name:
                return ParameterData(
                    name=str(self),
                    description="U component of wind at 100m above ground level. "
                                "Defined as the horizontal speed of "
                                "the wind in the eastward direction.",
                    units="m/s",
                    limits=ParameterLimits(upper=100, lower=-100),
                    alternative_shortnames=["u100"],
                    grib2_code="246",
                )

            case self.WIND_V_COMPONENT_100m.name:
                return ParameterData(
                    name=str(self),
                    description="V component of wind at 100m above ground level. "
                                "Defined as the horizontal speed of "
                                "the wind in the northward direction.",
                    units="m/s",
                    limits=ParameterLimits(upper=100, lower=-100),
                    alternative_shortnames=["v100"],
                    grib2_code="247",
                )

            case self.WIND_U_COMPONENT_200m.name:
                return ParameterData(
                    name=str(self),
                    description="U component of wind at 200m above ground level. "
                                "Defined as the horizontal speed of "
                                "the wind in the eastward direction.",
                    units="m/s",
                    limits=ParameterLimits(upper=150, lower=-150),
                    alternative_shortnames=["u200"],
                    grib2_code="239",
                )

            case self.WIND_V_COMPONENT_200m.name:
                return ParameterData(
                    name=str(self),
                    description="V component of wind at 200m above ground level. "
                                "Defined as the horizontal speed of "
                                "the wind in the northward direction.",
                    units="m/s",
                    limits=ParameterLimits(upper=150, lower=-150),
                    alternative_shortnames=["v200"],
                    grib2_code="240",
                )

            case self.WIND_DIRECTION_10m.name:
                return ParameterData(
                    name=str(self),
                    description="The wind direction from 0 to 360. 0 represents a Northerly "
                                "wind and 90 is Easterly wind. This is confirmed by the UK mean "
                                "wind direction being Westerly and = ~200. ",
                    units="degrees",
                    limits=ParameterLimits(upper=0, lower=360),
                    alternative_shortnames=["wdir", "wdir10", "10wdir"],
                    grib2_code="194",
                )

            case self.SNOW_DEPTH_GL.name:
                return ParameterData(
                    name=str(self),
                    description="Depth of snow on the ground.",
                    units="m",
                    limits=ParameterLimits(upper=12, lower=0),
                    alternative_shortnames=["sd", "sdwe", "sde"],
                    grib2_code="141",
                )

            case self.CLOUD_COVER_HIGH.name:
                return ParameterData(
                    name=str(self),
                    description="Fraction of grid square covered by high-level cloud. "
                                "Defined as the ratio of "
                                "the area of the grid square covered by high-level (>6km) cloud "
                                "to the square's total area.",
                    units="UI",
                    limits=ParameterLimits(upper=1, lower=0),
                    alternative_shortnames=["hcc"],
                    grib2_code="188",
                )

            case self.CLOUD_COVER_MEDIUM.name:
                return ParameterData(
                    name=str(self),
                    description="Fraction of grid square covered by medium-level cloud. "
                                "Defined as the ratio of "
                                "the area of the grid square covered by medium-level (2-6km) cloud "
                                "to the square's total area.",
                    units="UI",
                    limits=ParameterLimits(upper=1, lower=0),
                    alternative_shortnames=["mcc"],
                    grib2_code="187",
                )

            case self.CLOUD_COVER_LOW.name:
                return ParameterData(
                    name=str(self),
                    description="Fraction of grid square covered by low-level cloud. "
                                "Defined as the ratio of "
                                "the area of the grid square covered by low-level (<2km) cloud "
                                "to the square's total area.",
                    units="UI",
                    limits=ParameterLimits(upper=1, lower=0),
                    alternative_shortnames=["lcc"],
                    grib2_code="186",
                )

            case self.CLOUD_COVER_TOTAL.name:
                return ParameterData(
                    name=str(self),
                    description="Fraction of grid square covered by any cloud. "
                                "Defined as the ratio of "
                                "the area of the grid square covered by any cloud "
                                "to the square's total area.",
                    units="UI",
                    limits=ParameterLimits(upper=1, lower=0),
                    alternative_shortnames=["tcc", "clt"],
                    grib2_code="164",
                )

            case self.TOTAL_PRECIPITATION_RATE_GL.name:
                return ParameterData(
                    name=str(self),
                    description="Total precipitation rate at ground level. "
                                "Defined as the rate at which liquid is deposited on the ground "
                                "including rain, snow, and hail.",
                    units="kg/m^2/s",
                    limits=ParameterLimits(upper=0.2, lower=0),
                    alternative_shortnames=["prate", "tprate", "rprate"],
                    grib2_code="260048",
                )

            case self.DOWNWARD_ULTRAVIOLET_RADIATION_FLUX_GL.name:
                return ParameterData(
                    name=str(self),
                    description="Downward ultraviolet radiation flux at ground level. "
                                "Defined as the mean amount of "
                                "ultraviolet radiation incident on the surface "
                                "expected over the next hour.",
                    units="W/m^2",
                    limits=ParameterLimits(upper=1000, lower=0),
                    alternative_shortnames=["uvb"],
                    grib2_code="57",
                )

            case self.DIRECT_SHORTWAVE_RADIATION_FLUX_GL.name:
                return ParameterData(
                    name=str(self),
                    description="Direct shortwave radiation flux at ground level. "
                                "Defined as the mean amount of "
                                "unscattered solar radiation incident on"
                                "a surface plane perpendicular to the direction of the sun "
                                "expected over the next hour.",
                    units="W/m^2",
                    limits=ParameterLimits(upper=1000, lower=0),
                    alternative_shortnames=["dsrp"],
                    grib2_code="47",
                )

            case self.WIND_SPEED_10m.name:
                return ParameterData(
                    name=str(self),
                    description="Wind speed at 10m above ground level. "
                                "Defined as the magnitude of the wind vector.",
                    units="m/s",
                    limits=ParameterLimits(upper=150, lower=0),
                    alternative_shortnames=["10si", "si10"],
                    grib2_code="207",
                )

            case self.WIND_SPEED_100m.name:
                return ParameterData(
                    name=str(self),
                    description="Wind speed at 100m above ground level. "
                                "Defined as the magnitude of the wind vector.",
                    units="m/s",
                    limits=ParameterLimits(upper=200, lower=0),
                    alternative_shortnames=["100si", "si100"],
                    grib2_code="249",
                )

            case self.PRESSURE_MSL.name:
                return ParameterData(
                    name=str(self),
                    description="Mean sea level pressure. "
                                "Defined as the force per unit area of the atmosphere "
                                "adjusted to the height of mean sea level. This corresponds "
                                "to the weight of a column of air vertically above a point "
                                "on the Earth's surface. 100 Pa = 1 hPa = 1 mbar.",
                    units="Pa",
                    limits=ParameterLimits(upper=105000, lower=95000),
                    alternative_shortnames=["mslp", "msl"],
                    grib2_code="151",
                )
            case _:
                # Shouldn't happen thanks to the test case in test_parameters.py
                raise ValueError(f"Unknown parameter: {self}")

    def try_from_alternate(name: str) -> ResultE["Parameter"]:
        """Map an alternate name to a parameter."""
        for p in Parameter:
            if name in p.metadata().alternative_shortnames:
                return Success(p)
        return Failure(ValueError(f"Unknown shortname: {name}"))

    @staticmethod
    def rename_else_drop_ds_vars(
        ds: xr.Dataset, allowed_parameters: list["Parameter"],
    ) -> xr.Dataset:
        """Rename variables to match expected names, dropping invalid ones.

        Returns a dataset with all variables in it renamed to a known `entities.Parameter`
        name, if a matching parameter exists, and it is an allowed parameter. Otherwise,
        the variable is dropped from the dataset.

        Args:
            ds: The xarray dataset to rename.
            allowed_parameters: The list of parameters allowed in the resultant dataset.
        """
        for var in ds.data_vars:
            param_result = Parameter.try_from_alternate(str(var))
            match param_result:
                case Success(p):
                    if p in allowed_parameters:
                        ds = ds.rename_vars({var: p.value})
                        continue
            log.debug("Dropping invalid parameter '%s' from dataset", var)
            ds = ds.drop_vars(str(var))
        return ds




"""Dagster pipes notification repository implementation.

`Dagster Pipes <https://docs.dagster.io/_apidocs/pipes#dagster-pipes>`_
enables integration with Dagster for reporting asset materialization
and logging. This module enables dagster instances running this code to recieve
notifications.

See Also:
    - https://docs.dagster.io/concepts/dagster-pipes/subprocess/create-subprocess-asset
"""

import logging
from typing import override

from dagster_pipes import open_dagster_pipes
from returns.result import ResultE, Success

from nwp_consumer.internal import entities, ports

log = logging.getLogger("nwp-consumer")


class DagsterPipesNotificationRepository(ports.NotificationRepository):
    """Dagster pipes notification repository."""

    @override
    def notify(
        self,
        message: entities.StoreCreatedNotification | entities.StoreAppendedNotification,
    ) -> ResultE[str]:
        with open_dagster_pipes() as pipesctx:
            pipesctx.report_asset_materialization(
                metadata={
                    "filename": {"raw_value": message.filename, "type": "text"},
                    "size_mb": {"raw_value": message.size_mb, "type": "float"},
                    "memory_mb": {"raw_value": message.performance.memory_mb, "type": "float"},
                    "duration_minutes": {
                        "raw_value": int(message.performance.duration_seconds / 60),
                        "type": "int",
                    },
                },
            )
        return Success("Notification sent to dagster successfully.")




import dataclasses
import datetime as dt
import unittest

import pandas as pd
from returns.result import Failure, Success

from .coordinates import NWPDimensionCoordinateMap
from .parameters import Parameter


class TestCoordinates(unittest.TestCase):
    """Test the business methods of the NWPDimensionCoordinateMap class."""

    def test_determine_region(self) -> None:

        @dataclasses.dataclass
        class TestCase:
            name: str
            inner: NWPDimensionCoordinateMap
            expected_slices: dict[str, slice]
            should_error: bool

        outer: NWPDimensionCoordinateMap = NWPDimensionCoordinateMap(
            init_time=[dt.datetime(2021, 1, 1, i, tzinfo=dt.UTC) for i in range(0, 9, 3)],
            step=list(range(12)),
            variable=[
                Parameter.TEMPERATURE_SL,
                Parameter.CLOUD_COVER_HIGH,
                Parameter.TOTAL_PRECIPITATION_RATE_GL,
            ],
            latitude=[60.0, 61.0, 62.0],
            longitude=[10.0, 11.0, 12.0],
        )

        tests = [
            TestCase(
                name="basic_subset",
                inner=NWPDimensionCoordinateMap(
                    init_time=outer.init_time[:1],
                    step=outer.step[:6],
                    variable=outer.variable,
                    latitude=outer.latitude,
                    longitude=outer.longitude,
                ),
                expected_slices={
                    "init_time": slice(0, 1),
                    "step": slice(0, 6),
                    "variable": slice(0, 3),
                    "latitude": slice(0, 3),
                    "longitude": slice(0, 3),
                },
                should_error=False,
            ),
            TestCase(
                name="subset_with_multiple_span",
                inner=NWPDimensionCoordinateMap(
                    init_time=[
                        dt.datetime(2021, 1, 1, i, tzinfo=dt.UTC)
                        for i in [3, 6]
                    ],
                    step=outer.step,
                    variable=outer.variable,
                    latitude=outer.latitude,
                    longitude=outer.longitude,
                ),
                expected_slices={
                    "init_time": slice(1, 3),
                    "step": slice(0, 12),
                    "variable": slice(0, 3),
                    "latitude": slice(0, 3),
                    "longitude": slice(0, 3),
                },
                should_error=False,
            ),
            TestCase(
                name="subset_with_non_contiguous_values",
                inner=NWPDimensionCoordinateMap(
                    init_time=outer.init_time[:1],
                    step=list(range(1, 6, 2)),
                    variable=outer.variable,
                    latitude=[60.0, 63.0],
                    longitude=outer.longitude,
                ),
                expected_slices={},
                should_error=True,
            ),
            TestCase(
                name="not_a_subset",
                inner=NWPDimensionCoordinateMap(
                    init_time=outer.init_time[:1],
                    step=[15],
                    variable=outer.variable,
                    latitude=[12, 13, 14, 15],
                    longitude=outer.longitude,
                    ),
                expected_slices={},
                should_error=True,
            ),
            TestCase(
                name="different_dimensions",
                inner=NWPDimensionCoordinateMap(
                    init_time=outer.init_time[:1],
                    step=[15],
                    variable=outer.variable,
                ),
                expected_slices={},
                should_error=True,
            ),
        ]

        for t in tests:
            with self.subTest(name=t.name):
                result = outer.determine_region(inner=t.inner)
                if t.should_error:
                    self.assertTrue(
                        isinstance(result, Failure),
                        msg=f"{t.name}: Expected error to be returned.",
                    )
                else:
                    self.assertEqual(result, Success(t.expected_slices))

    def test_to_pandas(self) -> None:

        @dataclasses.dataclass
        class TestCase:
            name: str
            coords: NWPDimensionCoordinateMap
            expected_indexes: dict[str, pd.Index] # type: ignore

        tests = [
            TestCase(
                name="valid_data",
                coords=NWPDimensionCoordinateMap(
                    init_time=[dt.datetime(2021, 1, 1, i, tzinfo=dt.UTC) for i in range(0, 9, 3)],
                    step=list(range(12)),
                    variable=[
                        Parameter.CLOUD_COVER_HIGH,
                        Parameter.TEMPERATURE_SL,
                        Parameter.TOTAL_PRECIPITATION_RATE_GL,
                    ],
                    latitude=[60.0, 61.0, 62.0],
                    longitude=[10.0, 11.0, 12.0],
                ),
                expected_indexes={
                    "init_time": pd.to_datetime([
                        "2021-01-01T00:00:00Z",
                        "2021-01-01T03:00:00Z",
                        "2021-01-01T06:00:00Z",
                    ]),
                    "step": pd.Index([hour * 60 * 60 * 1000000000 for hour in range(12)]),
                    "variable": pd.Index([
                        Parameter.CLOUD_COVER_HIGH.value,
                        Parameter.TEMPERATURE_SL.value,
                        Parameter.TOTAL_PRECIPITATION_RATE_GL.value,
                    ]),
                    "latitude": pd.Index([62.0, 61.0, 60.0]),
                    "longitude": pd.Index([10.0, 11.0, 12.0]),
                },
            ),
        ]

        for t in tests:
            with self.subTest(name=t.name):
                result = t.coords.to_pandas()
                self.assertEqual(result["init_time"].dtype, "datetime64[ns]")
                self.assertListEqual(list(result.keys()), list(t.expected_indexes.keys()))
                for key in result:
                    self.assertListEqual(
                        result[key].values.tolist(),
                        t.expected_indexes[key].values.tolist())

    def test_from_pandas(self) -> None:
        @dataclasses.dataclass
        class TestCase:
            name: str
            data: dict[str, pd.Index]  # type: ignore
            expected_coordinates: NWPDimensionCoordinateMap | None
            should_error: bool

        tests = [
            TestCase(
                name="valid_data",
                data={
                    "init_time": pd.to_datetime(["2021-01-01T00:00:00Z", "2021-01-01T03:00:00Z"]),
                    "step": pd.to_timedelta(["0 days", "3 days"]),
                    "variable": pd.Index(["temperature_sl", "cloud_cover_high"]),
                    "latitude": pd.Index([61.0, 60.0]),
                    "longitude": pd.Index([10.0, 11.0]),
                },
                expected_coordinates=NWPDimensionCoordinateMap(
                    init_time=[
                        dt.datetime(2021, 1, 1, 0, tzinfo=dt.UTC),
                        dt.datetime(2021, 1, 1, 3, tzinfo=dt.UTC)],
                    step=[0, 72],
                    variable=[Parameter.TEMPERATURE_SL, Parameter.CLOUD_COVER_HIGH],
                    latitude=[60.0, 61.0],
                    longitude=[10.0, 11.0],
                ),
                should_error=False,
            ),
            TestCase(
                name="missing_required_keys",
                data={
                    "init_time": pd.to_datetime(["2021-01-01T00:00:00Z", "2021-01-01T03:00:00Z"]),
                    "step": pd.to_timedelta(["0 days", "3 days"]),
                    "latitude": pd.Index([61.0, 60.0]),
                    "longitude": pd.Index([10.0, 11.0]),
                },
                expected_coordinates=None,
                should_error=True,
            ),
            TestCase(
                name="unknown_parameter",
                data={
                    "init_time": pd.to_datetime(["2021-01-01T00:00:00Z", "2021-01-01T03:00:00Z"]),
                    "step": pd.to_timedelta(["0 hours", "1 hours", "2 hours", "3 hours"]),
                    "variable": pd.Index(["temperature_sl", "not_a_variable"]),
                    "latitude": pd.Index([61.0, 60.0]),
                    "longitude": pd.Index([10.0, 11.0], dtype="int64"),
                },
                expected_coordinates=None,
                should_error=True,
            ),
            TestCase(
                name="unknown_keys",
                data={
                    "init_time": pd.to_datetime(["2021-01-01T00:00:00Z", "2021-01-01T03:00:00Z"]),
                    "step": pd.to_timedelta(["0 days", "3 days"]),
                    "variable": pd.Index(["temperature_sl", "cloud_cover_high"]),
                    "latitude": pd.Index([61.0, 60.0]),
                    "longitude": pd.Index([10.0, 11.0]),
                    "unknown": pd.Index(["unknown"]),
                },
                expected_coordinates=None,
                should_error=True,
            ),
        ]

        for t in tests:
            with self.subTest(name=t.name):
                result = NWPDimensionCoordinateMap.from_pandas(t.data)
                if t.should_error:
                    self.assertTrue(
                        isinstance(result, Failure),
                        msg=f"{t.name}: Expected error to be returned.",
                    )
                else:
                    self.assertEqual(result, Success(t.expected_coordinates))

    def test_crop(self) -> None:
        """Test the crop method of the NWPDimensionCoordinateMap class."""

        @dataclasses.dataclass
        class Crop:
            n: float
            w: float
            s: float
            e: float

        @dataclasses.dataclass
        class TestCase:
            name: str
            coords: NWPDimensionCoordinateMap
            crop: Crop
            expected_latitude: list[float]
            expected_longitude: list[float]
            should_error: bool

        test_coords = NWPDimensionCoordinateMap(
            init_time=[dt.datetime(2021, 1, 1, i, tzinfo=dt.UTC) for i in range(0, 9, 3)],
            step=list(range(12)),
            variable=[
                Parameter.TEMPERATURE_SL,
                Parameter.CLOUD_COVER_HIGH,
                Parameter.TOTAL_PRECIPITATION_RATE_GL,
            ],
            latitude=[float(f"{lat / 10:.2f}") for lat in range(800, -800, -1)],
            longitude=[float(f"{lon / 10:.2f}") for lon in range(150, -150, -1)],
        )

        test_cases = [
            TestCase(
                name="basic_crop",
                coords=test_coords,
                crop=Crop(n=60.5, w=10.0, s=60, e=10.5),
                expected_latitude=[60.5, 60.4, 60.3, 60.2, 60.1, 60.0],
                expected_longitude=[10.0, 10.1, 10.2, 10.3, 10.4, 10.5],
                should_error=False,
            ),
            TestCase(
                name="crop_with_invalid_values",
                coords=test_coords,
                crop=Crop(n=60.0, w=10.0, s=61.0, e=12.0),
                expected_latitude=[60.0],
                expected_longitude=[10.0, 11.0],
                should_error=True,
            ),
            TestCase(
                name="crop_with_larger_values",
                coords=test_coords,
                crop=Crop(n=90.0, w=180, s=-90, e=-180),
                expected_latitude=test_coords.latitude if test_coords.latitude else [90.0],
                expected_longitude=test_coords.longitude if test_coords.longitude else [180.0],
                should_error=True,
            ),
        ]

        for t in test_cases:
            with self.subTest(name=t.name):
                result = t.coords.crop(
                    north=t.crop.n,
                    west=t.crop.w,
                    south=t.crop.s,
                    east=t.crop.e,
                )
                if t.should_error:
                    self.assertIsInstance(result, Failure)
                else:
                    self.assertIsInstance(result, Success)
                    coords = result.unwrap()
                    self.assertListEqual(coords.latitude, t.expected_latitude) # type: ignore
                    self.assertListEqual(coords.longitude, t.expected_longitude) # type: ignore


if __name__ == "__main__":
    unittest.main()




import dataclasses
import datetime as dt
import os
import pathlib
import unittest
from unittest.mock import patch

from returns.result import Failure, ResultE, Success

from nwp_consumer.internal import entities

from ...entities import NWPDimensionCoordinateMap
from .ceda import CEDARawRepository


@dataclasses.dataclass
class TestCase:
    filename: str
    should_error: bool
    expected_coords: entities.NWPDimensionCoordinateMap

class TestCEDARawRepository(unittest.TestCase):
    """Test the business methods of the CEDAFTPRawRepository class."""

    def test__convert_global(self) -> None:
        """Test the _convert_global method."""


        tests: list[TestCase] = [
            TestCase(
                filename="test_CEDAFTP_UM-Global_ssrd_20241105T00_S01-03.grib",
                expected_coords = dataclasses.replace(
                    entities.Models.MO_UM_GLOBAL_17KM.expected_coordinates,
                    init_time=[dt.datetime(2024, 11, 5, 0, tzinfo=dt.UTC)],
                    step=[1, 2, 3],
                    variable=[entities.parameters.Parameter.DOWNWARD_SHORTWAVE_RADIATION_FLUX_GL],
                ),
                should_error=False,
            ),
            TestCase(
                filename="test_CEDAFTP_UM-Global_u_20241105T00_S01-03_AreaC.grib",
                expected_coords = dataclasses.replace(
                    entities.Models.MO_UM_GLOBAL_17KM.expected_coordinates,
                    init_time=[dt.datetime(2024, 11, 5, 0, tzinfo=dt.UTC)],
                    step=[1, 2, 3],
                    variable=[entities.parameters.Parameter.WIND_U_COMPONENT_10m],
                ),
                should_error=False,
            ),
            TestCase(
                filename="test_MODatahub_UM-Global_t2m_20241120T00_S00.grib",
                expected_coords = entities.Models.MO_UM_GLOBAL_17KM.expected_coordinates,
                should_error=True,
            ),
        ]


        for t in tests:
            with self.subTest(name=t.filename):
                # Attempt to convert the file
                result = CEDARawRepository._convert_global(
                    pathlib.Path(__file__).parent.absolute() / "test_gribs" / t.filename,
                )
                region_result: ResultE[dict[str, slice]] = result.do(
                    region
                    for das in result
                    for da in das
                    for region in NWPDimensionCoordinateMap.from_xarray(da).bind(
                        t.expected_coords.determine_region,
                    )
                )
                if t.should_error:
                    self.assertIsInstance(region_result, Failure, msg=f"{region_result}")
                else:
                    self.assertIsInstance(region_result, Success, msg=f"{region_result}")

    @patch.dict(os.environ, {"MODEL": "mo-um-ukv"}, clear=True)
    def test__convert_ukv(self) -> None:
        """Test the _convert_ukv method."""
        tests: list[TestCase] = [
            TestCase(
                filename="test_subset_Wholesale1.grib",
                expected_coords=dataclasses.replace(
                    entities.Models.MO_UM_UKV_2KM_OSGB.expected_coordinates,
                    init_time=[dt.datetime(2022, 1, 1, 3, tzinfo=dt.UTC)],
                    step=[1, 2],
                    variable=[entities.parameters.Parameter.TEMPERATURE_SL],
                ),
                should_error=False,
            ),
            TestCase(
                filename="test_subset_Wholesale2.grib",
                expected_coords=dataclasses.replace(
                    entities.Models.MO_UM_UKV_2KM_OSGB.expected_coordinates,
                    init_time=[dt.datetime(2022, 1, 1, 3, tzinfo=dt.UTC)],
                    step=[1, 2],
                    variable=[entities.parameters.Parameter.CLOUD_COVER_LOW],
                ),
                should_error=False,
            ),
        ]

        for t in tests:
            with self.subTest(name=t.filename):
                # Attempt to convert the file
                result = CEDARawRepository._convert_ukv(
                    pathlib.Path(__file__).parent.absolute() / "test_gribs" / t.filename,
                )
                region_result: ResultE[dict[str, slice]] = result.do(
                    region
                    for das in result
                    for da in das
                    for region in NWPDimensionCoordinateMap.from_xarray(da).bind(
                        t.expected_coords.determine_region,
                    )
                )
                if t.should_error:
                    self.assertIsInstance(region_result, Failure, msg=f"{region_result}")
                else:
                    self.assertIsInstance(region_result, Success, msg=f"{region_result}")

if __name__ == "__main__":
    unittest.main()
